package net.surguy.logfilter;

import org.objectweb.asm.*;

import java.lang.instrument.*;
import java.lang.reflect.*;
import java.security.ProtectionDomain;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.io.IOException;
import java.io.FileOutputStream;

/**
 * Preloader using JDK 1.5 instrumentation to completely remove Log4J calls and the overhead of methods called within
 * them. Call with the JVM parameter:
 * <p>
 * <code>
 *  -javaagent:net.surguy.logfilter.LogPreloader=debug
 * </code>
 * <p>
 * where the option can be one of "debug", "info", "warn", "error" or "fatal", and defaults to "debug" if not set.
 *
 * @author Inigo Surguy, inigosurguy@hotmail.com
 * @created 08-Feb-2004 09:24:08
 */
public class LogPreloader implements ClassFileTransformer {

    public static void premain(String options, Instrumentation instrumentation) {
        LogLevel highestLevelToRemove;
        try {
            highestLevelToRemove = (options == null) ? LogLevel.debug : LogLevel.valueOf(options);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("Option passed to LogPreloader class transformer must be one of " + Arrays.asList(LogLevel.values()));
        }
        instrumentation.addTransformer(new LogPreloader(highestLevelToRemove));
    }

    private final LogLevel highestLevelToRemove;
    private LogPreloader(LogLevel highestLevelToRemove) {
        this.highestLevelToRemove = highestLevelToRemove;
    }

    public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer) throws IllegalClassFormatException {
        // Returning null means that no transformation was done.
        if (isSystemClass(className) || (isLog4JClass(className))) { return null; }
        ClassReader reader = new ClassReader(classfileBuffer);
        final int majorVersion = 49;
        final int minorVersion = 0;
        ClassWriter writer = new ClassWriter(true, majorVersion, minorVersion);
        ClassVisitor adapter = new RemoveLoggingClassAdapter(writer, highestLevelToRemove);
        reader.accept(adapter, false);
        byte[] results = writer.toByteArray();
        return results;
    }


    private boolean isLog4JClass(String className) {
        return className.startsWith("org/apache/log4j");
    }

    private boolean isSystemClass(String className) {
        return ((className.startsWith("java/")) || (className.startsWith("javax/")) || (className.startsWith("sun/")));
    }
}

enum LogLevel { debug, info, warn, error, fatal };

class RemoveLoggingClassAdapter extends ClassAdapter implements Constants {
    private final LogLevel highestLevelToRemove;

    public RemoveLoggingClassAdapter(ClassVisitor cv, LogLevel highestLevelToRemove) {
        super(cv);
        this.highestLevelToRemove = highestLevelToRemove;
    }

    public CodeVisitor visitMethod(int access, String name, String desc, String[] exceptions, Attribute attrs) {
        CodeVisitor methodVisitor = cv.visitMethod(access, name, desc, exceptions, attrs);
        CodeVisitor dynamicLogRemover = (CodeVisitor) Proxy.newProxyInstance(this.getClass().getClassLoader(),
                                   new Class[]{CodeVisitor.class}, new DynamicLogRemovingHandler(methodVisitor, highestLevelToRemove));
        return methodVisitor == null ? null : dynamicLogRemover;
    }

    public void visitInnerClass(String name, String outerName, String innerName, int access) {
        super.visitInnerClass(name, outerName, innerName, access);    //To change body of overridden methods use File | Settings | File Templates.
    }
}

class DynamicLogRemovingHandler implements InvocationHandler {

    // Lazily created, because methods without calls to the logger (probably a high proportion) won't require it
    private List<Pair<Method,Object[]>> savedInstructions;
    private boolean isStoring = false;
    private final CodeVisitor codeVisitor;
    private final LogLevel highestLevelToRemove;

    public DynamicLogRemovingHandler(CodeVisitor methodVisitor, LogLevel highestLevelToRemove) {
        this.codeVisitor = methodVisitor;
        this.highestLevelToRemove = highestLevelToRemove;
    }

    /**
     * Most ASM calls will be passed through directly to the wrapped codeVisitor -
     */
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        if (method.getName().equals("visitFieldInsn")) {
            final Object accessedFieldClass = args[3];
            if ("Lorg/apache/log4j/Category;".equals(accessedFieldClass)) { isStoring = true; }
        } else if (method.getName().equals("visitMethodInsn")) {
            final Object calledObjectName = args[1];
            final String calledMethodName = (String) args[2];
            if (("org/apache/log4j/Category".equals(calledObjectName))) {
                if ((isLogMethod(calledMethodName)) && (LogLevel.valueOf(calledMethodName).compareTo(highestLevelToRemove)<=0)) {
                    isStoring = false;
                    clearSavedInstructions();
                    return null;
                } else {
                    isStoring = false;
                    replaySavedInstructions();
                    clearSavedInstructions();
                }
            }
        } else if (method.getName().equals("visitMaxs")) {
            isStoring = false;
            replaySavedInstructions();
            clearSavedInstructions();
        }

        if (isStoring) {
            store(new Pair<Method, Object[]>(method, args));
        } else {
            method.invoke(codeVisitor, args);
        }
        return null;
    }

    private boolean isLogMethod(String methodName) {
        for (LogLevel level : LogLevel.values())
            if (level.toString().equals(methodName)) return true;
        return false;
    }

    private void replaySavedInstructions() throws IllegalAccessException, InvocationTargetException {
        if (savedInstructions==null) { return; }
        for (Pair<Method, Object[]> instruction : savedInstructions) {
            instruction.item1.invoke(codeVisitor, instruction.item2);
        }
    }

    private void clearSavedInstructions() {
        if (savedInstructions!=null) { savedInstructions.clear(); }
    }

    private void store(Pair<Method, Object[]> methodToStore) {
        if (savedInstructions==null) { savedInstructions = new ArrayList<Pair<Method,Object[]>>(); }
        savedInstructions.add(methodToStore);
    }

}

class Pair<Item1, Item2> {
    final Item1 item1;
    final Item2 item2;
    public Pair(Item1 item1, Item2 item2) {
        this.item1 = item1;
        this.item2 = item2;
    }
}
